! Good GPU heat equation solver in 2D.

program heat_solve_gpu_good
    use heat
    use core_gpu_good
    use setup
    use utilities
    use omp_lib

    implicit none

    real(dp), parameter :: a = 0.5      ! Diffusion constant.
    type(field) :: current, previous    ! Current and previus temperature fields.
    real(dp) :: dt                      ! Time step.
    integer :: nsteps                   ! Number of time steps.

    integer :: iter                     ! Counter of iterations for loop.
    real(dp) :: average_temp            ! Average temperature
    real(kind=dp) :: start, stop        ! Timers

    ! Create copies of the field variables used by the evolve function.
    integer :: cur_nx, cur_ny
    real(dp) :: cur_dx, cur_dy
    real(dp), dimension(:,:), allocatable :: prev_data, cur_data

    ! Initialize the field.
    call initialize(current, previous, nsteps)
    average_temp = average(current)
    write(*, '(A,F9.6)') 'Average temperature at start: ', average_temp

    ! Initialize the copies.
    cur_nx = current%nx
    cur_ny = current%ny
    cur_dx = current%dx
    cur_dy = current%dy
    allocate(prev_data(0:previous%nx+1, 0:previous%ny+1))
    allocate(cur_data(0:cur_nx+1, 0:cur_ny+1))
    prev_data = previous%data
    cur_data = current%data

    ! Largest stable time step
    dt = current%dx**2 * current%dy**2 / &
         & (2.0 * a * (current%dx**2 + current%dy**2))

    ! Main iteration loop, save a picture every
    ! image_interval steps
    start =  omp_get_wtime()

    !$omp target data map(to:cur_nx,cur_ny,cur_dx,cur_dy,dt,prev_data) map(tofrom:cur_data)

    do iter = 1, nsteps
       call evolve(cur_nx, cur_ny, cur_dx, cur_dy, prev_data, cur_data, a, dt)
       !$omp target
       prev_data = cur_data
       !$omp end target
    end do

    !$omp end target data

    stop = omp_get_wtime()

    ! Copy results to field object for finalization.
    previous%data = prev_data

    ! Average temperature for reference
    average_temp = average(previous)

    write(*, '(A30,F6.3,A)') 'Iterations took: ', stop - start, '    seconds.'
    write(*, '(A30,F9.6)') 'Average temperature: ',  average_temp
    if (command_argument_count() == 0) then
        write(*, '(A30,F9.6)') 'Reference value for default: ', 59.281239
    end if

    deallocate(prev_data)
    deallocate(cur_data)
    call finalize(current, previous)

end program heat_solve_gpu_good
